
"""PURPOSE: launch the mini-editor to create custom p setups"""

import glob
import os
import stat
import subprocess

from sc2gameLobby import gameConstants as c
from sc2gameLobby.gameConfig import Config

# windows 10
banksDir = "C:\\Users\\jared\\Documents\\StarCraft II\\Banks"
availableBankNames = glob.glob(os.path.join(banksDir, "*"))
editorCmd = "%s -run %s -testMod %s -displayMode 1"
#EXAMPLE:
#C:\\StarCraft II\\Support64>SC2Switcher_x64.exe -run "Ladder\\CatalystLE.SC2Map" -testMod "C:\\Users\\TheWisp\\Documents\\PlayGround.SC2Mod"

################################################################################
def launchEditor(mapObj):
    """
    PURPOSE: launch the editor using a specific map object
    INPUT:   mapObj (sc2maptool.mapRecord.MapRecord)
    """
    cfg = Config()
    if cfg.is64bit: selectedArchitecture = c.SUPPORT_64_BIT_TERMS
    else:           selectedArchitecture = c.SUPPORT_32_BIT_TERMS
    fullAppPath = os.path.join(
        cfg.installedApp.data_dir,
        c.FOLDER_APP_SUPPORT%(selectedArchitecture[0]))
    appCmd = c.FILE_EDITOR_APP%(selectedArchitecture[1])
    fullAppFile = os.path.join(fullAppPath, appCmd)
    os.chmod(fullAppFile, stat.S_IRUSR|stat.S_IRGRP|stat.S_IROTH|stat.S_IXUSR|\
                          stat.S_IRUSR|stat.S_IWUSR|stat.S_IWGRP|stat.S_IXGRP) # switcher file must be executable
    finalCmd = editorCmd%(appCmd, mapObj.path, c.FILE_EDITOR_MOD)
    cwd = os.getcwd()
    os.chdir(fullAppPath)
    os.system(finalCmd)
    os.chdir(cwd) # restore back to original directory





from s2clientprotocol import sc2api_pb2
from pysc2.lib import protocol
from pysc2.lib import remote_controller

import portpicker
import queue
import socket
import sys
import time
import websocket

from sc2gameLobby import gameConstants as c


################################################################################
class ClientController(remote_controller.RemoteController):
    """similar to pysc2 StarcratProcess, but without the process to enable multiple game connections"""
    ############################################################################
    def __init__(self, url=None, port=None, timeout=c.INITIAL_TIMEOUT):
        sys.argv = sys.argv[:1] # trim to force flags to do nothing
        FLAGS = protocol.flags.FLAGS
        FLAGS(sys.argv) # always ensure flags did its checking
        self._url       = None
        self._port      = None
        self._client    = None
        self._name      = ""
        if url!=None or port!=None:
            self.connect(url=url, port=port, timeout=timeout)
    ############################################################################
    def __str__(self): return self.__repr__()
    def __repr__(self):
        url  = " %s"     %self._url  if self._url  else ""
        port = ":%s"%self._port if self._port else ""
        try:    stat = self.status
        except: stat = "disconnected"
        return "<%s%s%s %s>"%(self.name, url, port, stat)
    ############################################################################
    def __nonzero__(self):
        """whether this ClientController is connected"""
        try:    self.status
        except: return False
        return True
    ############################################################################
    @property
    def name(self):
        if not self._name: # execute only once, if needed
            self._name = str(self.__class__).split('.')[-1].rstrip("'>")
        return self._name
    ############################################################################
    def close(self):
        """Shut down the socket connection, client and controller"""
        self._sock = None
        self._controller = None
        if hasattr(self, "_port") and self._port:
            portpicker.return_port(self._port)
            self._port = None
    ############################################################################
    def __enter__(self):
        return self
    ############################################################################
    def __exit__(self, unused_exception_type, unused_exc_value, unused_traceback):
        self.close()
    ############################################################################
    def __del__(self):
        # Prefer using a context manager, but this cleans most other cases.
        self.close()
    ############################################################################
    def connect(self, url=c.LOCALHOST, port=None, timeout=c.INITIAL_TIMEOUT,
                      debug=False):
        """socket connect to an already running starcraft2 process"""
        if port != None: # force a selection to a new port
            if self._port!=None: # if previously allocated port, return it
                portpicker.return_port(self._port)
            self._port = port
        elif self._port==None: # no connection exists
            self._port = portpicker.pick_unused_port()
        self._url = url
        if ":" in url and not url.startswith("["):  # Support ipv6 addresses.
            url = "[%s]" % url
        for i in range(timeout):
            startTime = time.time()
            if debug:
                print("attempt #%d to websocket connect to %s:%s"%(i, url, port))
            try:
                finalUrl = "ws://%s:%s/sc2api" %(url, self._port)
                ws = websocket.create_connection(finalUrl, timeout=timeout)
                #print("ws:", ws)
                self._client = protocol.StarcraftProtocol(ws)
                #super(ClientController, self).__init__(client) # ensure RemoteController initializtion is performed
                #if self.ping(): print("init ping()") # ensure the latest state is synced
                # ping returns:
                #   game_version:   "4.1.2.60604"
                #   data_version:   "33D9FE28909573253B7FC352CE7AEA40"
                #   data_build:     60604
                #   base_build:     60321
                return self
            except socket.error: pass  # SC2 hasn't started listening yet.
            except websocket.WebSocketException as err:
                print(err, type(err))
                if "Handshake Status 404" in str(err):
                    pass  # SC2 is listening, but hasn't set up the /sc2api endpoint yet.
                else: raise
            except Exception as e:
                print(type(e), e)
            sleepTime = max(0, 1 - (time.time() - startTime)) # try to wait for up to 1 second total
            if sleepTime:   time.sleep(sleepTime)
        raise websocket.WebSocketException("Could not connect to game at %s on port %s"%(url, port))
    ############################################################################
    def getNewReplay(self):
        request = sc2api_pb2.RequestReplayInfo(replay_path="dummy.SC2Replay", download_data=True)
        #request.replay_path = ?
        #request.replay_data = ?
        #request.download_data=True
        return self._client.send(replay_info=request)


ClientController.__bool__ = ClientController.__nonzero__ # python2-3 compatibility




"""
PURPOSE: send request communications to a ladder server

    sendMatchResult()

    requestMatch()
        find matchup
        listen for matchup assignment
        launch client using assignment info
        issue commands as assigned
        play game
            if W, agent app result code  = 0
            else  agent app result code != 0
        reportResult() to ladder

    export / publish data
        standings by W/L
        standings by games played / sportsmanship rating
        hottest streak in standings w/ 5+ games
        service getPlayerData() request (provide info about a ladder's players)
"""

from six import iteritems # python 2/3 compatibility

import json
import requests

from sc2gameLobby import gameConstants as c
from sc2players import addPlayer, getPlayer


################################################################################
def cancelMatchRequest(cfg):
    """obtain information housed on the ladder about playerName"""
    payload = json.dumps([cfg.thePlayer])
    ladder = cfg.ladder
    return requests.post(
        url  = c.URL_BASE%(ladder.ipAddress, ladder.serverPort, "cancelmatch"),
        data = payload,
        #headers=headers,
    )


################################################################################
def ladderPlayerInfo(cfg, playerName, getMatchHistory=False):
    """obtain information housed on the ladder about playerName"""
    payload = json.dumps([playerName, getMatchHistory]) # if playerName == None, info on all players is retrieved
    ladder = cfg.ladder
    return requests.post(
        url  = c.URL_BASE%(ladder.ipAddress, ladder.serverPort, "playerstats"),
        data = payload,
        #headers=headers,
    )


################################################################################
def reportMatchCompletion(cfg, results, replayData):
    """send information back to the server about the match's winners/losers"""
    payload = json.dumps([cfg.flatten(), results, replayData])
    ladder = cfg.ladder
    return requests.post(
        url  = c.URL_BASE%(ladder.ipAddress, ladder.serverPort, "matchfinished"),
        data = payload,
        #headers=headers,
    )


################################################################################
def sendMatchRequest(cfg):
    payload = cfg.toJson()
    ladder = cfg.ladder
    return requests.post(
        # force error #url  = c.URL_BASE%(ladder.ipAddress, ladder.serverPort, "polls"),
        url  = c.URL_BASE%(ladder.ipAddress, ladder.serverPort, "newmatch"),
        data = payload,
        #headers=headers,
    )
    #print("x", x)
    #print("dir(x)", dir(x))
    #print("x.raw", x.raw)
    #print("x.url", x.url)
#    print("x.text", x.text) # this holds the response content
#    print("x.ok", x.ok)
#    print("x.reason", x.reason)




"""
Copyright (c) 2018 Versentiedge LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS-IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from sc2gameLobby.cli import main

if __name__=="__main__":
    main()




"""
platform.system()
    Linux: Linux
    Mac: Darwin
    Windows: Windows

can change the windows server process name (PyPi modules)
    wmi-client-wrapper (0.0.12)      - Linux-only wrapper around wmi-client for WMI (Windows)
    WMI (1.4.9)                      - Windows Management Instrumentation
    wmi-client-wrapper-py3 (2018.3)  - A fork of Bryan Bishop's wmi-client-wrapper extended to include Python3 support

https://opensource.com/resources/python/gui-frameworks


> wrap/utilize game lobby / config


        ** causes unofficial, custom game
            desired map name
            desired opponent playerID
            fog of war
            realtime or step rate



    UI to select options
        select game critera
            desiredRace
            agent to play
            generated comm

"""


import platform




"""
Copyright (c) 2018 Versentiedge LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS-IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import absolute_import
from __future__ import division  # python 2/3 compatibility
from __future__ import print_function  # python 2/3 compatibility

import os

from sc2gameLobby import gameConfig
from sc2gameLobby import gameConstants
from sc2gameLobby import getGameData
from sc2gameLobby import genericObservation as go
from sc2gameLobby import hostGame
from sc2gameLobby import joinGame
from sc2gameLobby import versions
from sc2gameLobby.__version__ import *

config = gameConfig.Config  # set up the player/match environment as desired
host = hostGame.run  # host a new Starcraft2 match, given a configuration
join = joinGame.playerJoin  # join an existing Starcraft2 match, given a configuration
clear = gameConfig.clearConfigs  # manually reset/clear any existing (defunct) game configurations
active = gameConfig.activeConfigs  # display the games currently being set up


def updateVersion(**kwargs):
    """add/update record data using kwargs params for new keys/values"""
    versions.handle.save(kwargs)




from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

from six import iteritems # python 2/3 compatibility
from io import BytesIO # python 2/3 compatibility

from sc2gameLobby import dateFormat
from sc2gameLobby import gameConstants as c

import json
import os
import re
import time


"""
Starcraft 2 patch/update information is copied from the following site(s):
    http://liquipedia.net/starcraft2/Patch_4.2.3 (and links listed on the page)
"""


################################################################################
def addNew(label, version, baseVersion, dataHash="", fixedHash="", replayHash=""):
    """
    Add a new version record to the database to be tracked
    VERSION RECORD EXAMPLE:
        "base-version": 55505, 
        "data-hash": "60718A7CA50D0DF42987A30CF87BCB80", 
        "fixed-hash": "0189B2804E2F6BA4C4591222089E63B2", 
        "label": "3.16", 
        "replay-hash": "B11811B13F0C85C29C5D4597BD4BA5A4", 
        "version": 55505
        """
    baseVersion = int(baseVersion)
    version     = int(version)
    minVersChecks = {"base-version":baseVersion, "version":version}
    if label in handle.ALL_VERS_DATA:
        raise ValueError("given record label (%s) is already defined.  Consider performing update() for this record instead"%(label))
    for vCheckK,vCheckV in iteritems(minVersChecks): # verify no conflicting values
        maxVersion  = min([vData[vCheckK] for vData in handle.ALL_VERS_DATA.values()])
        if vCheckV < c.MIN_VERSION_AI_API:
            raise ValueError("version %s / %s.%s does not support the Starcraft2 API"%(baseVersion, label, version))
        if vCheckV < maxVersion: # base version cannot be smaller than newest value
            raise ValueError("given %s (%d) cannot be smaller than newest known %s (%d)"%(vCheckK, vCheckV, vCheckK, maxVersion))
    uniqueValHeaders = list(c.JSON_HEADERS)
    uniqueValHeaders.remove("base-version")
    record = {"base-version" : baseVersion}
    #print("%15s : %s (%s)"%("base-version", baseVersion, type(baseVersion)))
    for k,v in zip(uniqueValHeaders, [label, version, dataHash, fixedHash, replayHash]): # new attr values must be unique within all handler records
        record[k] = v # convert to dict while checking each param
        #print("%15s : %s (%s)"%(k,v,type(v)))
        if not v: continue # ignore uniqueness requirement if value is unspecified
        if v in [r[k] for r in Handler.ALL_VERS_DATA.values()]:
            raise ValueError("'%s' '%s' is in known values: %s"%(k, v, getattr(handle, k)))
            return
    handle.save(new=record)


################################################################################
#def update(version, newValues):
#    """
#    USAGE EXAMPLES:
#        update(59729, {"fixed-hash":"123"})
#        update("3.19", {"replay-hash":"WHATEVER231"})
#    """
#    if len(newValues) == 0: raise ValueError("must provide data in paramater newValues else cannot update any record")
#    try:
#            int(version)
#            versionKey = "version"
#    except: versionKey = "label"
#    records = handle.search(version) # match record where versionKey value == version
#    if   len(records) > 1: raise ValueError("identified too many records (%d): %s"%(len(records), records))
#    elif len(records) < 1: raise ValueError("failed to identify any records given params: %s"%(records))
#    record = records.pop() # access the single, found record
#    for k,v in iteritems(newValues):   record[k] = v
#    handle._updated = True
#    handle.save()


################################################################################
class Handler(object):
    """NOTE: the data-hash field is required to launch stand-alone versions of the game"""
    ############################################################################
    ALL_VERS_DATA = None
    ############################################################################
    def __init__(self):
        self.load()
    ############################################################################
    #def __getattr__(self, key):
    #    """get all record values of a given key"""
    #    return sorted([v[key] for v in Handler.ALL_VERS_DATA.values()])
    ############################################################################
    def __len__(self):
        """the number of known version records"""
        return len(Handler.ALL_VERS_DATA)
    ############################################################################
    @property
    def mostRecent(self):
        records = iteritems(Handler.ALL_VERS_DATA)
        try: 
            label,record = max(records)
            return record
        except ValueError: return None # no versions are installed
    ############################################################################
    #def get(self, key, versFilter=""):
    #    """get a values of key where records match versFilter"""
    #    for v in self.search(versFilter): print(v)
    #    return [v[key] for v in self.search(versFilter)]
    ############################################################################
    def load(self):
        """load ALL_VERS_DATA from disk"""
        basepath = os.path.dirname(os.path.abspath(__file__))
        filename = os.sep.join([basepath, c.FOLDER_JSON, c.FILE_GAME_VERSIONS])
        Handler.ALL_VERS_DATA = {} # reset known data; do not retain defunct information
        with open(filename, "r") as f:
            data = json.loads( f.read() )
        self.update(data)
        self._updated = False
        #for v,record in iteritems(Handler.ALL_VERS_DATA):
        #    print(type(v), v)
            #for k,v in iteritems(record):    
    ############################################################################
    def save(self, new=None, timeout=2):
        """write ALL_VERS_DATA to disk in 'pretty' format"""
        if new: self.update(new) # allow two operations (update + save) with a single command
        if not self._updated: return # nothing to do
        thisPkg = os.path.dirname(__file__)
        filename = os.path.join(thisPkg, c.FOLDER_JSON, c.FILE_GAME_VERSIONS)
        fParts = c.FILE_GAME_VERSIONS.split('.')
        newFile = os.path.join(thisPkg, c.FOLDER_JSON, "%s_%s.%s"%(fParts[0], dateFormat.now(), fParts[1]))
        if not os.path.isfile(newFile):
            #fParts = c.FILE_GAME_VERSIONS.split('.')
            #newFile = "%s%s%s_%s.%s"%(c.FOLDER_JSON, os.sep, fParts[0], dateFormat.now(), fParts[1])
            #if not os.path.isfile(newFile):
            #print(filename)
            #print(newFile)
            os.rename(filename, newFile) # backup existing version file
        recordKeys = [(record["version"], record) for record in Handler.ALL_VERS_DATA.values()]
        data = [r for k,r in sorted(recordKeys)] # i.e. get values sorted by version key
        start = time.time()
        while time.time()-start < timeout: # allow multiple retries if multiple processes fight over the version file
            try:
                with open(filename, "wb") as f:
                    f.write(str.encode(json.dumps(data, indent=4, sort_keys=True))) # python3 requires encoding str => bytes to write to file
                self._updated = False
                return
            except IOError: pass # continue waiting for file to be available
        raise # after timeout, prior exception is what matters
    ############################################################################
    def search(self, *args, **kwargs):
        """match all records that have any args in any key/field that also match
        key/value requirements specified in kwargs"""
        ret = []
        for record in Handler.ALL_VERS_DATA.values():
            matchArgs = list(kwargs.keys())
            for k,v in iteritems(kwargs): # restrict records based on key-value match requirement
                try:
                    if record[k] != v: break # a non-matching requirement means this record doesn't match
                except: break # record doesn't have required key 'k'
                matchArgs.remove(k)
            if matchArgs: continue # didn't match all required kwargs
            matchArgs = list(args)
            for k,v in iteritems(record): # find any record with a <value> in it
                if k in matchArgs: matchArgs.remove(k)
                if v in matchArgs: matchArgs.remove(v)
            if matchArgs: continue # didn't match all required args
            ret.append(record)
        return ret
    ############################################################################
    #def search(self, searchVal, keyOnly=None, **filters):
    #    """identify all version records where searchVal exists (can be a regex)"""
    #    ########################################################################
    #    def checkValues(vData, regex):
    #        for k,v in iteritems(vData):
    #            #if k == "base-version": continue # ignore matching this key; expect it will match 'version' key where applicable
    #            if re.search(regex, str(v)): return True
    #        return False
    #    ########################################################################
    #    _tmp = re.compile("")
    #    if type(searchVal) == type(_tmp):
    #          regex = searchVal
    #    else: regex = re.compile(str(searchVal))
    #    ret = {}
    #    if keyOnly!=False: # search all keys (only)
    #        for vers in Handler.ALL_VERS_DATA.values():
    #            if any([re.search(regex, str(k)) for k in vers.keys()]):
    #                ret[vers["label"]] = vers
    #    if keyOnly!=True: # search all values (only)
    #        for vers in Handler.ALL_VERS_DATA.values():
    #            if checkValues(vers, regex):   ret[vers["label"]] = vers
    #    return ret.values()
    ############################################################################
    def update(self, data):
        """update known data with with newly provided data"""
        if not isinstance(data, list): data = [data] # otherwise no conversion is necessary
        master = Handler.ALL_VERS_DATA
        for record in data:
            #print(record)
            for k,v in iteritems(record): # ensure record contents aretyped appropriately
                try:                record[k] = int(v)
                except ValueError:  record[k] = v
            try: label = record["label"] # verify this record has the required 'label' key
            except KeyError:
                raise ValueError("Must provide a valid label argument.  Given:%s%s"%(\
                    os.linesep, ("%s  "%(os.linesep)).join(
                        ["%15s:%s"%(k,v) for k,v in iteritems(kwargs)]
                    )))
            try:    masterLabel = master[label] # identify the already existing record that matches this to-be-updated record, if any
            except KeyError: # master hasn't been defined yet
                master[label] = record
                self._updated = True # a new record should also be saved
                continue
            for k,v in iteritems(record): # determine whether master needs to be updated
                try:
                    if masterLabel[k] == v:  continue # whether an entry in the record needs to be updated (doesn't match)
                except KeyError:             pass # this condition means that k is a new key, so the record must be updated
                self._updated = True
                try:    master[label].update(record) # index each record by its label
                except KeyError:             break


################################################################################
handle = Handler()
#newRecord = {
#    "fixed-hash": "", 
#    "replay-hash": "", 
#    "data-hash": "", 
#    "label": "4.2.4", 
#    "base-version": 63454, 
#    "version": 64128
#}
#handle.save(new=newRecord)


################################################################################
class Version(object):
    """Represent a single version of the game settings"""
    ############################################################################
    def __init__(self, versionVal=None):
        if versionVal: # get specific version
            if isinstance(versionVal, list): versionVal = max(versionVal)
            records = handle.search(versionVal)
            if   len(records) > 1: raise ValueError("identified too many records (%d): %s"%(len(records), records))
            elif len(records) < 1: raise ValueError("first collect and update version information for version: %s"%(versionVal))
            record = records.pop()
        else: # select most recent version
            record = handle.mostRecent
        for k in c.JSON_HEADERS: # convert hyphenated keys into code-compatible keys
            oldK = k
            wordBoundaries = re.search("(-\w)", k)
            if wordBoundaries:
                for wb in wordBoundaries.groups():
                    newWb = wb.upper().strip('-')
                    k = re.sub(wb, newWb, k)
            setattr(self, k, record[oldK])
        while self.label.count(".") < 2: self.label += ".0" # ensure a 3-field versioning scheme
    ############################################################################
    def __str__(self):  return self.__repr__()
    def __repr__(self):
        return "%s.%s"%(self.label, self.version)
    ############################################################################
    def __getitem__(self, key):
        return getattr(self, key)
    ############################################################################
    def toTuple(self, typeCast=int):
        return [typeCast(v) for v in self.label.split('.')]
    ############################################################################
    def toFilename(self):
        return re.sub("\.", "_", str(self.label))





from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

import json
import socket
import time
try:                from urllib2        import urlopen
except Exception:   from urllib.request import urlopen

from sc2gameLobby import gameConstants as c

'''
SOURCES:
    http://stackoverflow.com/questions/9481419/how-can-i-get-the-public-ip-using-python2-7
'''


################################################################################
def getAll():
    return (getPublicIPaddress(), getLocalIPaddress(), getMachineIPaddress())


################################################################################
def getMachineIPaddress():
    """visible on this local machine only"""
    return c.LOCALHOST


################################################################################
def getLocalIPaddress():
    """visible to other machines on LAN"""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('google.com', 0))
        my_local_ip = s.getsockname()[0] # takes ~0.005s
        #from netifaces import interfaces, ifaddresses, AF_INET
        #full solution in the event of multiple NICs (network interface cards) on the PC
        #def ip4_addresses():
        #    ip_list = []
        #    for interface in interfaces():
        #        for link in ifaddresses(interface)[AF_INET]: # If IPv6 addresses are needed instead, use AF_INET6 instead of AF_INET
        #            ip_list.append(link['addr'])
        #    return ip_list
    except Exception:
        my_local_ip = None
    return my_local_ip # would need to change change if using multiple NICs


################################################################################
def getPublicIPaddress(timeout=c.DEFAULT_TIMEOUT):
    """visible on public internet"""
    start = time.time()
    my_public_ip = None
    e = Exception
    while my_public_ip == None:
        if time.time() - start > timeout:
            break
        try: #httpbin.org -- site is useful to test scripts / applications.
            my_public_ip = json.load(urlopen('http://httpbin.org/ip'))['origin'] # takes ~0.14s as ipv4
            if my_public_ip: break
        except Exception as e:
            print(type(e), e, "http://httpbin.org/ip")
        try: #jsonip.com -- Seemingly the sole purpose of this domain is to return IP address in JSON.
            my_public_ip = json.load(urlopen('http://jsonip.com'))['ip']  # takes ~0.24s as ipv6
            if my_public_ip: break
        except Exception as e:
            print(type(e), e, "http://jsonip.com")
        try: #ipify.org -- Power of this service results from lack of limits (there is no rate limiting), infrastructure (placed on Heroku, with high availability in mind) and flexibility (works for both IPv4 and IPv6).
            my_public_ip = load(urlopen('https://api.ipify.org/?format=json'))['ip']  # takes ~0.33s
            if my_public_ip: break
        except Exception as e:
            print(type(e), e, "https://api.ipify.org/")
        try: #ip.42.pl -- This is very convenient for scripts, you don't need JSON parsing here.
            my_public_ip = urlopen('http://ip.42.pl/raw').read()  # takes ~0.35s
            if my_public_ip: break
        except Exception as e:
            print(type(e), e, "http://ip.42.pl/raw")
    if not my_public_ip:
        raise e
    return my_public_ip





from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

import os

from sc2common.constants import *
from sc2common import types

################################################################################
class TimeoutExceeded(Exception): pass
class UnknownPlayer(Exception): pass

################################################################################
# file/folder information
SC2_FILE_REPLAY         = "SC2Replay"
SC2_FILE_MAP            = "SC2Map"
FOLDER_LOBBY_HERE       = os.path.dirname(os.path.abspath(__file__))
FOLDER_PLAYED_VIDEO     = "playedReplays"
FOLDER_JSON             = "jsonData"
FOLDER_ACTIVE_CONFIGS   = os.path.join(FOLDER_LOBBY_HERE, "activeGames")
FOLDER_MODS             = os.path.join(FOLDER_LOBBY_HERE, "mods")
JSON_HEADERS            = ["label", "base-version", "version", "data-hash", "fixed-hash", "replay-hash"]
FILE_GAME_VERSIONS      = "versions.json"
FILE_EDITOR_MOD         = os.path.join(FOLDER_MODS, "Playground.SC2Mod")
MIN_VERSION_AI_API      = 55958 # 3.16.1 is the first version that released the API
FOLDER_IGNORED_MAPS     = ["Melee", "mini_games", "Test"]
FOLDER_APP_SUPPORT      = "Support%s"         # reserved space to specify 32-bit or 64-bit items
FILE_EDITOR_APP         = "SC2Switcher%s.exe" # reserved space to specify 32-bit or 64-bit items
SUPPORT_32_BIT_TERMS    = ["", ""]
SUPPORT_64_BIT_TERMS    = ["64", "_x64"]

################################################################################
# misc
MIN_REQUIRED_PLAYERS    = 2   # players
INITIAL_TIMEOUT         = 120 # seconds
DEFAULT_TIMEOUT         = 15  # seconds
DEFAULT_HOST_DELAY      = 4   # seconds
REPLAY_SAVE_FREQUENCY   = 10  # seconds
URL_BASE                = "http://%s:%s/%s/"





from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

from pysc2.lib import protocol
from pysc2.lib import remote_controller
from pysc2.lib.sc_process import FLAGS

import base64
import os
import subprocess
import sys
import time

from sc2gameLobby.clientManagement import ClientController
from sc2gameLobby import gameConfig
from sc2gameLobby import gameConstants as c
from sc2gameLobby import resultHandler as rh

now = time.time


################################################################################
def run(config, agentCallBack, lobbyTimeout=c.INITIAL_TIMEOUT, debug=True):
    """PURPOSE: start a starcraft2 process using the defined the config parameters"""
    FLAGS(sys.argv) # ignore pysc2 command-line handling (eww)
    log = protocol.logging.logging
    log.disable(log.CRITICAL) # disable pysc2 logging
    amHosting   = not bool(config.host)
    thisPlayer  = config.whoAmI()
    operPrefix  = "HOST" if amHosting else "JOIN"
    operType    = "%sGAME"%operPrefix
    createReq   = config.requestCreateDetails() if amHosting else None
    joinReq     = config.requestJoinDetails()
    selectedIP  = config.clientInitHost()
    selectPort  = config.clientInitPort()
    controller  = None # the object that manages the application process
    finalResult = rh.playerSurrendered(config) # default to this player losing if somehow a result wasn't acquired normally
    replayData  = "" # complete raw replay data for the match
    if debug: print("[%s] Starcraft2 game process is launching (fullscreen=%s)."%(operType, config.fullscreen))
    with config.launchApp(ip_address=selectedIP, port=selectPort, connect=False):
      try: # WARNING: if port equals the same port of the host on the same machine, this subsequent process closes!
        controller = ClientController()
        controller.connect(url=selectedIP, port=selectPort, timeout=lobbyTimeout) # establish socket connection
        if amHosting:
            if debug:
                  print("[%s] Starcraft2 host application is live. (%s)"%(operType, controller.status)) # status: launched
                  print("[%s] Creating Starcraft Game at %s"%(operType, controller))
            controller.create_game(createReq)
            if debug:
                  print("[%s] Starcraft2 is waiting for %d player(s) to join. (%s)"%(operType, config.numAgents, controller.status)) # status: init_game
                  print("[%s] sending request to join game. (%s)"%(operType, controller.status)) # status: init_game
        else: # joining clients must wait for host to perfrom its join request
            timeToWait = c.DEFAULT_HOST_DELAY
            for i in range(timeToWait): # WARNING: the host must perform its join action with its client before any joining players issue join requests to their clients
                if debug: print("[%s] waiting %d seconds for the host to finish its init sequence."%(operType, timeToWait-i))
                time.sleep(1)
        joinResp = controller.join_game(joinReq) # SC2APIProtocol.RequestJoinGame
        print("[%s] connection to %s:%d was successful. Game is starting! (%s)"%(operType, selectedIP, selectPort, controller.status)) # status: in_game
        thisPlayer.playerID = int(joinResp.player_id) # update playerID; repsponse to join game request is authority
        if debug: print("[%s] joined match as %s."%(operType, thisPlayer)) # all players have actually joined already to advance beyond join_game (init_game)
        config.updateIDs(controller.game_info(), tag=operType, debug=debug) # SC2APIProtocol.ResponseGameInfo object
        if debug: print("[%s] all %d player(s) found; game has started! (%s)"%(operType, config.numGameClients, controller.status)) # status: init_game
        config.save() # "publish" the configuration file for other procs
        try:    agentCallBack(config.name) # send the configuration to the controlling agent
        except Exception as e:
            print("ERROR: agent %s crashed during init: %s (%s)"%(thisPlayer.initCmd, e, type(e)))
            return (rh.playerCrashed(config), "") # no replay information to get
#        
#        ### added
#        if amHosting and config.startCondition:
#              simulationSetup(controller, config.startCondition)
#        else: timeToWait = 0
#        
        getGameState = controller.observe # function that observes what's changed since the prior gameloop(s)
        startWaitTime = now()
#        
#        ### added
#        scenarioStart = startWaitTime
#        
        while True:  # wait for game to end while players/bots do their thing
            obs = getGameState()
            result = obs.player_result
            if result: # match end condition was supplied by the client
                finalResult = rh.idPlayerResults(config, result)
                break
            try: agentCallBack(obs) # do developer's creative stuff
            except Exception as e:
                print("%s ERROR: agent callback %s of %s crashed during game: %s"%(type(e), agentCallBack, thisPlayer.initCmd, e))
                finalResult = rh.playerCrashed(config)
                break
            newNow = now() # periodicially acquire the game's replay data (in case of abnormal termination)
#            
#            ### added
#            if newNow - scenarioStart > timeToWait: # scenario time has elapsed
#                testCase = config.setup
#                if not simulationSetup(config.startConditions)
#            
            if newNow - startWaitTime > c.REPLAY_SAVE_FREQUENCY:
                replayData = controller.save_replay()
                startWaitTime = newNow
        replayData = controller.save_replay() # one final attempt to get the complete replay data
        #controller.leave() # the connection to the server process is (cleanly) severed
      except (protocol.ConnectionError, protocol.ProtocolError, remote_controller.RequestError) as e:
        if "Status.in_game" in str(e): # state was previously in game and then exited that state
            finalResult = rh.playerSurrendered(config) # rage quit is losing
        else:
            finalResult = rh.playerDisconnected(config)
            print("%s Connection to game host has ended, even intentionally by agent. Message:%s%s"%(type(e), os.linesep, e))
      except KeyboardInterrupt:
        if debug: print("caught command to forcibly shutdown Starcraft2 client.")
        finalResult = rh.playerSurrendered(config)
      finally:
        if replayData: # ensure replay data can be transmitted over http
            replayData = base64.encodestring(replayData).decode() # convert raw bytes into str
        if controller: controller.quit() # force the sc2 application to close
    return (finalResult, replayData)





from __future__ import absolute_import
from __future__ import division       # python 2/3 compatibility
from __future__ import print_function # python 2/3 compatibility

import datetime

################################################################################
def now(givenTime=None):
    if givenTime:   usableTime = datetime.datetime.utcfromtimestamp(givenTime)
    else:           usableTime = datetime.datetime.utcnow()
    now = usableTime.replace(microsecond=0)
    now = now.isoformat("_").replace(":", "_")
    return now





import sys

import sc2gameLobby

from sc2gameLobby import cli

def test_import():
    assert bool(sc2gameLobby)
    

#def test_cli():
#    sys.argv = sys.argv[:1] + ["--search=defaulthuman"]
#    cli.main() # good request
#    sys.argv = sys.argv[:1] + ["--search=playerDoesNotExist4235nxprdjhqwhmdfxu"]
#    cli.main() # bad request
#    #sys.argv = sys.argv[:1] + ["--nogui"]
#    #cli.main() # NYI -- requires auto-dependency installation
#    sys.argv = sys.argv[:1] + ["--add"]
#    cli.main()
#    sys.argv = sys.argv[:1] + ["--update"]
#    cli.main()
#    sys.argv = sys.argv[:1] + ["--versions"]
#    cli.main()
#    sys.argv = sys.argv[:1] + ["--windowed", "--raw"]
#    cli.main() # launch GUI attempt, but error
#    sys.argv = sys.argv[:1] # restore sys.argv to no-arg state


#def test_cli_error1():
#    """test an invalid/offline server configuration"""
#    sys.argv = sys.argv[:1] + ["--ladder=testInvalid"]
#    cli.main()


#def test_cli_error2():
#    """test an invalid/offline server configuration"""
#    sys.argv = sys.argv[:1] + ["--ladder=testInvalid"]
#    cli.main()
